/*--------------------------------------------------------------------
This source distribution is placed in the public domain by its author,
Jason Papadopoulos. You may use it for any purpose, free of charge,
without having to notify anyone. I disclaim any responsibility for any
errors.

Optionally, please be nice and tell me if you find this source to be
useful. Again optionally, if you add to the functionality present here
please consider making those additions public too, so that others may 
benefit from your work.	
       				   --jasonp@boo.net 6/3/07
--------------------------------------------------------------------*/

#include "poly.h"

#if MAX_POLY_DEGREE < 6
#error "Polynomial generation assumes degree <= 6 allowed"
#endif

/* main structure for representing the factors of
   the leading rational poly coefficient */

typedef struct {
	uint16 prime;
	uint16 id;
	uint16 roots[MAX_POLY_DEGREE];
} rat_factor_t;

#define MAX_RAT_FACTORS 80
#define MAX_CURR_RAT_FACTORS 15

/* Leading coefficients of the algebraic polynomial 
   are found by sieving */

#define ALG_SIEVE_SIZE 4096

/* we sieve with powers of primes. Each power gets 
   the same log value, and log values stack up 
   at locations divisible by powers of a prime */

static const uint8 alg_sieve_p[] = 
	{8, 8, 8, 9, 9, 9, 5, 5, 5, 5, 7, 7, 7, 11, 11,
	13, 13, 17, 17, 19, 19, 23, 23, 29, 29, 31, 31};

/* the sieving stride used in each prime power */

static const uint16 alg_sieve_powers[] = 
	{8, 64, 512, 9, 81, 729, 5, 25, 125, 625, 
	 7, 49, 343, 11, 121, 13, 169, 17, 289, 
	 19, 361, 23, 529, 29, 841, 31, 961};

#define NUM_ALG_SIEVE_ENTRIES sizeof(alg_sieve_p)

/* main structure controlling leading algebraic 
   coefficient sieving */

typedef struct {
	uint32 strides[NUM_ALG_SIEVE_ENTRIES];
	uint32 roots[NUM_ALG_SIEVE_ENTRIES];
	uint8 logs[NUM_ALG_SIEVE_ENTRIES];
	uint8 cutoff;
	uint8 *sieve;
} alg_sieve_t;

/* scale factor for logs of sieve primes */

#define ALG_SIEVE_LOG_SCALE 6.0

/*------------------------------------------------------------------*/
static void alg_sieve_init(uint64 base, alg_sieve_t *s) {

	/* Sieving will begin from 'base' and proceed in 
	   blocks of size ALG_SIEVE_SIZE */

	uint32 i;
	double dbase = (double)base;

	s->sieve = (uint8 *)aligned_malloc(ALG_SIEVE_SIZE, 16);
	s->cutoff = (uint8)(ALG_SIEVE_LOG_SCALE * log(dbase) / M_LN2 + 0.5);
	
	for (i = 0; i < NUM_ALG_SIEVE_ENTRIES; i++) {
		uint32 root;
		s->strides[i] = alg_sieve_powers[i];
		s->logs[i] = (uint8)(ALG_SIEVE_LOG_SCALE * 
				log((double)alg_sieve_p[i]) / 
				M_LN2 + 0.5);
		root = base % s->strides[i];
		if (root > 0)
			root = s->strides[i] - root;
		s->roots[i] = root;
	}

	if (s->cutoff > s->logs[i-1])
		s->cutoff -= s->logs[i-1];
	else
		s->cutoff = 0;
}

/*------------------------------------------------------------------*/
static void alg_sieve_free(alg_sieve_t *s) {
	aligned_free(s->sieve);
}

/*------------------------------------------------------------------*/
static void alg_sieve_update_logs(uint64 new_base, alg_sieve_t *s) {

	/* as sieve values grow, it becomes harder and harder
	   to find smooth values. Update the sieving cutoff 
	   to reflect this */

	double dbase = (double)(new_base);

	s->cutoff = (uint8)(ALG_SIEVE_LOG_SCALE * 
				log(dbase) / M_LN2 + 0.5);

	if (s->cutoff > s->logs[NUM_ALG_SIEVE_ENTRIES-1])
		s->cutoff -= s->logs[NUM_ALG_SIEVE_ENTRIES-1];
	else
		s->cutoff = 0;
}

/*------------------------------------------------------------------*/
static void alg_sieve_next(alg_sieve_t *s) {
	uint32 i;
	uint8 *sieve = s->sieve;

	/* fill the next sieve block. Sieve values begin
	   with the target log value and are decremented
	   as primes update the block. Values that meet
	   the cutoff get their sign bit set, which is
	   easy to check in parallel */

	memset(sieve, s->cutoff, (size_t)ALG_SIEVE_SIZE);
	for (i = 0; i < NUM_ALG_SIEVE_ENTRIES; i++) {
		uint32 root = s->roots[i];
		uint32 stride = s->strides[i];
		uint8 logp = s->logs[i];

		while (root < ALG_SIEVE_SIZE) {
			sieve[root] -= logp;
			root += stride;
		}
		s->roots[i] = root - ALG_SIEVE_SIZE;
	}
}

/*------------------------------------------------------------------*/
static void save_poly(mp_t *n, poly_config_t *config,
		      mp_t *m, mp_t *high_coeff) {

	uint32 i;
	mp_poly_t rpoly;
	mp_poly_t apoly;
	
	memset(&rpoly, 0, sizeof(rpoly));
	memset(&apoly, 0, sizeof(apoly));

	/* form the rational polynomial */

	rpoly.degree = 1;
	mp_copy(m, &rpoly.coeff[0].num);
	rpoly.coeff[0].sign = NEGATIVE;
	rpoly.coeff[1].num.nwords = 1;
	rpoly.coeff[1].num.val[0] = 1;
	rpoly.coeff[1].sign = POSITIVE;

	/* perform the analysis */

	analyze_poly(config, &rpoly, &apoly);
}

/*------------------------------------------------------------------*/
void poly_skew_stage2(mp_t *n, uint32 degree,
			poly_config_t *config, 
			rat_factor_t *rat_factors, 
			uint32 num_rat_factors, 
			mp_t *m0_in, uint64 alg_coeff) {

	uint32 i, j;
	mp_t r1, r0;
	mp_t m0;
	mp_t crt_tmp[MAX_CURR_RAT_FACTORS];

	/* compute r1, the high-order rational coefficient */

	mp_clear(&r1);
	r1.nwords = 1;
	r1.val[0] = rat_factors[0].prime;
	for (i = 1; i < num_rat_factors; i++)
		mp_mul_1(&r1, rat_factors[i].prime, &r1);

	/* compute the auxiliary quantities needed for CRT
	   reconstruction of r0 */

	for (i = 0; i < num_rat_factors; i++) {

		uint32 p = rat_factors[i].prime;

		mp_divrem_1(&r1, p, crt_tmp + i);
		j = mp_mod_1(crt_tmp + i, p);
		j = mp_modinv_1(j, p);
		mp_mul_1(crt_tmp + i, j, crt_tmp + i);
	}

	/* round m0 up to the next multiple of r1 */

	mp_mod(m0_in, &r1, &m0);
	mp_sub(&r1, &m0, &m0);
	mp_add(m0_in, &m0, &m0);

}

/*------------------------------------------------------------------*/
void poly_skew_stage1(mp_t *n, uint32 degree,
			poly_config_t *config, uint16 *factor_list, 
			uint32 num_rat_factors, uint64 alg_coeff) {

	/* after the leading algebraic poly coefficient is
	   chosen we have to choose the leading rational
	   coefficient. We do so in bulk here */

	uint32 i, j, k;
	uint32 max_rat_factors;
	uint32 max_curr_rat_factors;
	rat_factor_t rat_factors[MAX_CURR_RAT_FACTORS];
	rat_factor_t curr_rat_factors[MAX_CURR_RAT_FACTORS];
	mp_t m0, tmp1, tmp2;

	/* determine how many factors the leading rational
	   coefficient will have. This depends on the size 
	   of the leading algebraic coefficient; we want the
	   product of the primes chosen to be larger, but
	   not too much larger, than the algebraic coefficient.
	   This is because the second hghest algebraic coefficient
	   is bounded in size by the leading rational coefficient */

	if (alg_coeff < 1.0e4)
		max_curr_rat_factors = 5;
	else if (alg_coeff < 1.0e6)
		max_curr_rat_factors = 6;
	else if (alg_coeff < 1.0e8)
		max_curr_rat_factors = 7;
	else if (alg_coeff < 1.0e10)
		max_curr_rat_factors = 8;
	else
		max_curr_rat_factors = 9;

	/* determine the size of the pool of primes from
	   which the factors of the rational coefficient
	   are chosen. A smaller number of factors needs a 
	   larger pool, since one group of factors will
	   generate fewer polynomials in that case */

	switch (max_curr_rat_factors) {
	case 5: max_rat_factors = MAX_CURR_RAT_FACTORS; break;
	case 6: max_rat_factors = MAX_CURR_RAT_FACTORS - 1; break;
	case 7: max_rat_factors = MAX_CURR_RAT_FACTORS - 2; break;
	case 8: max_rat_factors = MAX_CURR_RAT_FACTORS - 2; break;
	default: max_rat_factors = MAX_CURR_RAT_FACTORS - 3; break;
	}

	/* In order to generate valid NFS polynomial pairs, 
	   we have to choose the rational polynomial R(x) = 
	   R1 * x - R0. This routine picks R1 and 
	   poly_skew_stage2() picks R0. If the leading 
	   algebraic coefficient is A we must have 
	   
	   	N == A*R0^degree mod R1   (*)
	   
	   R1 will be the product of several small primes; 
	   a prime p is suitable for use within R1 if (*) has
	   one or more solutions mod p. That way we can pick
	   several p and use the Chinese Remainder Thm. to solve
	   for R0. We want p to have many solutions to (*), so
	   we can generate many R0 for a given collection of 
	   primes p, to give more chances to find a good algebraic
	   polynomial. 
	   
	   For degree d a prime p has at most d solutions to (*),
	   and this can happen only if p % d == 1.  However, a given
	   p that satisfies this will have either d solutions or
	   zero solutions to (*). The loop below creates a pool of
	   primes p that have d solutions to (*). */

	for (i = j = 0; i < num_rat_factors; i++) {
		uint32 p = factor_list[i];
		rat_factor_t *curr_factor = rat_factors + j;
		uint32 nmodp = mp_mod_1(n, p);
		uint32 amodp = alg_coeff % p;
		uint32 count = 0;

		if (amodp == 0)
			continue;

		amodp = mp_modinv_1(amodp, p);
		nmodp = mp_modmul_1(nmodp, amodp, p);

		/* find the d roots of (*) mod p, if they
		   exist, and save them for later */

		curr_factor->prime = p;
		for (k = 1; k < p; k++) {
			if (mp_expo_1(k, degree, p) == nmodp) {
				curr_factor->roots[count] = k;
				if (++count == degree)
					break;
			}
		}

		if (count == degree) {
			/* make curr_factor remember that it is
			   entry j in the list */
			curr_factor->id = j;
			if (++j == max_rat_factors)
				break;
		}
	}

	/* we cannot pick more p at once than there 
	   are entries in the pool */

	max_rat_factors = j;
	max_curr_rat_factors = MIN(max_curr_rat_factors, 
				   max_rat_factors);
	if (max_rat_factors < 3)
		return;

	/* compute m0 ( = (N/A)^(1/degree) ), the starting 
	   value for the other rational coefficient R0. Each 
	   distinct R1 will modify m0 slightly to get a 
	   suitable R0. */

	mp_clear(&tmp1);
	tmp1.val[0] = (uint32)alg_coeff;
	tmp1.val[1] = (uint32)(alg_coeff >> 32);
	tmp1.nwords = 1;
	if (tmp1.val[1])
		tmp1.nwords = 2;

	mp_div(n, &tmp1, &tmp2);
	mp_iroot(&tmp2, NULL, degree, &m0);

	/* now enumerate all the collections of i primes in the
	   pool, and generate the polynomials for each collection.
	   We also process all the collections with one prime less
	   than was specified. The enumeration algorithm is a wonderful
	   combinatorial method called NEXKSB by Nijenhuis and Wilf */

	for (i = max_curr_rat_factors - 1; 
			i <= max_curr_rat_factors; i++) {

		uint32 first_entry = 0;
		uint32 list_size = i;

		do {
			/* assemble the primes that have changed from
			   the last iteration */

			for (j = 0; j < list_size; j++) {
				curr_rat_factors[i + j - list_size] =
						rat_factors[first_entry + j];
			}

			/* generate NFS polynomials */

			poly_skew_stage2(n, degree, config, 
					curr_rat_factors, i, &m0,
					alg_coeff);

			/* figure out how many entries have to change
			   in the next iteration */

			if (first_entry < max_rat_factors - list_size)
				list_size = 0;
			list_size++;

			/* figure out the oridinal number of the entries
			   they will change into */

			first_entry = curr_rat_factors[i - list_size].id + 1;

			/* enumeration stops when the last i-tuple of
			   primes is loaded */
		} while (curr_rat_factors[0].id != max_rat_factors - i);
	}
}

/*------------------------------------------------------------------*/
#define PACKED_SIEVE_MASK ((uint64)0x80808080 << 32 | 0x80808080)

void find_poly_skew(msieve_obj *obj, mp_t *n,
			poly_config_t *config,
			uint32 deadline) {

	/* search for NFS polynomials */

	uint32 i, j, k;
	uint32 default_cofactor;
	alg_sieve_t alg_sieve;
	uint64 alg_coeff_min = 1;
	uint32 blocks_done;
	time_t start_time, curr_time;
	uint16 rat_factors[MAX_RAT_FACTORS];
	uint32 num_rat_factors;
	uint32 degree;

	/* all leading algebraic coefficients will be 
	   divisible by default_cofactor. This is chosen 
	   both to ease the sieving burden and to give 
	   all polynomials more projective roots for the 
	   small primes */

	default_cofactor = 2 * 2 * 3 * 5;

	alg_sieve_init(alg_coeff_min, &alg_sieve);

	/* choose the polynomial degree */

	i = mp_bits(n);
	if (i < 347)
		degree = 4;	/* up to 105 digits */
	else if (i < 627)
		degree = 5;	/* up to 190 digits */
	else
		degree = 6;

	/* the leading rational poly coefficient will 
	   be the product of several small primes. A prime
	   p is put in the list if p mod degree == 1,
	   meaning there are p / degree values modulo p
	   that have a degree_th root modulo p */

	for (i = 3, num_rat_factors = 0; 
				i < config->prime_list.num_primes; i++) {
		uint32 p = config->prime_list.list[i];
		if (p % degree == 1) {
			rat_factors[num_rat_factors] = p;
			if (++num_rat_factors == MAX_RAT_FACTORS)
				break;
		}
	}

	blocks_done = 0;
	time(&start_time);
	while (1) {

		uint64 *packed_sieve = (uint64 *)alg_sieve.sieve;

		/* run the sieve, looking for numbers with many
		   small factors */

		alg_sieve_next(&alg_sieve);

		for (j = 0; j < ALG_SIEVE_SIZE/8; j += 4) {
			if (((packed_sieve[j+0] | 
			      packed_sieve[j+1] |
			      packed_sieve[j+2] | 
			      packed_sieve[j+3]) & 
			      PACKED_SIEVE_MASK) == (uint64)0)
				continue;

			for (k = 0; k < 32; k++) {
				if (alg_sieve.sieve[8 * j + k] > 
						alg_sieve.cutoff) {

					/* generate all of the polynomials
					   that have a given leading algebraic 
					   coefficient */

					poly_skew_stage1(n, degree,
							config, rat_factors, 
							num_rat_factors,
							default_cofactor *
							(alg_coeff_min + 
								(8 * j + k)));
				}
			}
		}
		alg_coeff_min += ALG_SIEVE_SIZE;
		alg_sieve_update_logs(alg_coeff_min, &alg_sieve);

		/* quit if interrupted */

		if (obj->flags & MSIEVE_FLAG_STOP_SIEVING)
			break;

		/* print progress message, quit if time expired */

		if (++blocks_done % 1024 == 0) {
			time(&curr_time);
			fprintf(stderr, "%u%% done (processed %u blocks)\r", 
				MIN(100, (uint32)(100.0 * difftime(
				    curr_time, start_time) / deadline + 0.5)),
				blocks_done);
			fflush(stderr);
			if (difftime(curr_time, start_time) > deadline)
				break;
		}
	}

	fprintf(stderr, "\n");
	alg_sieve_free(&alg_sieve);
}
