/*--------------------------------------------------------------------
This source distribution is placed in the public domain by its author,
Jason Papadopoulos. You may use it for any purpose, free of charge,
without having to notify anyone. I disclaim any responsibility for any
errors.

Optionally, please be nice and tell me if you find this source to be
useful. Again optionally, if you add to the functionality present here
please consider making those additions public too, so that others may 
benefit from your work.	
       				   --jasonp@boo.net 6/3/07
--------------------------------------------------------------------*/

#include "poly.h"
#include <gsl/gsl_errno.h>

typedef struct {
	uint32 degree;
	dd_t coeff[MAX_POLY_DEGREE + 1];
} dd_poly_t;

typedef struct {
	dd_poly_t rpoly;
	dd_poly_t apoly;
} integrate_t;

/*------------------------------------------------------------------*/
static double integrand(double x, void *params) {

	/* callback for numerical integration */

	integrate_t *info = (integrate_t *)params;
	dd_poly_t *rpoly = &info->rpoly;
	dd_poly_t *apoly = &info->apoly;
	dd_t rval, aval;
	int32 degree = rpoly->degree + apoly->degree;
	uint32 i;

	/* evaluate each polynomial in extended precision
	   using Horner's rule */

	i = rpoly->degree;
	rval = rpoly->coeff[i];
	while (i) {
		rval = dd_mul_d(rval, x);
		rval = dd_add_dd(rval, rpoly->coeff[--i]);
	}

	i = apoly->degree;
	aval = apoly->coeff[i];
	while (i) {
		aval = dd_mul_d(aval, x);
		aval = dd_add_dd(aval, apoly->coeff[--i]);
	}

	/* compute the integrand using only the high-order 
	   double of rval and aval */

	return pow(fabs(rval.hi), -2.0 / degree) * 
	       pow(fabs(aval.hi), -2.0 / degree);
}

/*------------------------------------------------------------------*/
static void dd_poly_mp2dd(dd_poly_t *dd_poly, mp_poly_t *mp_poly) {

	uint32 i;

	dd_poly->degree = mp_poly->degree;
	for (i = mp_poly->degree; (int32)i >= 0; i--) {
		dd_t coeff = dd_mp2dd(&mp_poly->coeff[i].num);
		if (mp_poly->coeff[i].sign == NEGATIVE) {
			coeff.hi = -coeff.hi;
			coeff.lo = -coeff.lo;
		}
		dd_poly->coeff[i] = coeff;
	}
}

/*------------------------------------------------------------------*/
static uint32 analyze_poly_size(poly_config_t *config,
				mp_poly_t *rpoly, mp_poly_t *apoly, 
				double *result) {

	gsl_function F;
	integrate_t info;
	double size_error;

	F.function = &integrand;
	F.params = &info;
	memset(&info, 0, sizeof(info));
	dd_poly_mp2dd(&info.rpoly, rpoly);
	dd_poly_mp2dd(&info.apoly, apoly);

	/* Rather than use Murphy's method to rate the average
	   size of polynomial values, we use a result of D.J.
	   Bernstein: For rational poly R(x) and algebraic poly
	   A(x), the number of polynomial values where x is rational
	   and the size of A(x)*R(x) achieves a given value is 
	   proportional to the integral of

	   	dx / ((R(x) * A(x))^2) ^ (1/(deg(R(x))+deg(A(x))))

	   for x from -infinity to +infinity. Larger values of
	   this integral imply more values that get found by
	   a siever. Bernstein writes that this estimate is 
	   "extremely accurate", but has not elaborated.

	   We use a black box numerical integrator to compute
	   this integral. */
	   
	return gsl_integration_qagi(&F, 0.0, 1.0e-6, (size_t)1000, 
				config->integrate_aux, result, 
				&size_error);
}

/*------------------------------------------------------------------*/
#define NUM_SMALL_PRIMES 27

static double analyze_poly_roots(poly_config_t *config, mp_poly_t *poly) {

	/* analyze a polynomial for root properties (i.e.
	   compute Murphy's 'alpha' value). For primes under 100
	   this uses the divisibility properties of a moderately
	   large set of polynomial sample values, and for larger 
	   primes we just count the roots that would appear in a 
	   factor base */

	uint32 i, j;
	double root_score;
	uint32 small_counts[NUM_SMALL_PRIMES];

	/* all linear polynomials have the same root properties */

	if (poly->degree == 1)
		return 0.0;

	/* experimentally determine how often the smallest primes
	   occur in a random sample of (a,b) pairs */

	memset(small_counts, 0, sizeof(small_counts));

	for (i = 0; i < NUM_POLY_SAMPLES; i++) {
		abpair_t *sample = config->samples + i;
		signed_mp_t sample_val;

		eval_poly(&sample_val, sample->a, sample->b, poly);

		for (j = 0; j < NUM_SMALL_PRIMES; j++) {
			uint32 prime = config->prime_list.list[j];
			uint32 num_factors = 0;
			while (mp_mod_1(&sample_val.num, prime) == 0) {
				num_factors++;
				mp_divrem_1(&sample_val.num, prime,
						&sample_val.num);
			}
			small_counts[j] += num_factors;
		}
	}

	/* compute the average contribution of the small primes */

	root_score = 0;
	for (i = 0; i < NUM_SMALL_PRIMES; i++) {
		uint32 prime = config->prime_list.list[i];
		root_score += log((double)prime) * 
				(1.0 / (prime - 1) - 
				 (double)small_counts[i] / NUM_POLY_SAMPLES);
	}

	/* for the larger primes, use Murphy's estimate for
	   the average contribution */

	for (; i < config->prime_list.num_primes; i++) {
		uint32 prime = config->prime_list.list[i];
		uint32 num_roots;
		uint32 dummy_roots[MAX_POLY_DEGREE];
		uint32 high_coeff;

		num_roots = poly_get_zeros(dummy_roots, poly, 
					prime, &high_coeff, 1);
		if (high_coeff == 0)
			num_roots++;

		root_score += (1.0 - (double)num_roots * prime / 
					(prime + 1)) * log((double)prime) / 
					(prime - 1);
	}

	return root_score;
}

/*------------------------------------------------------------------*/
void analyze_poly(poly_config_t *config,
		      mp_poly_t *rpoly, mp_poly_t *apoly) {

	/* analyze a polynomial for sieving goodness.
	   If the polynomial is good enough, save it.
	  
	   The analysis routines are general enough so that
	   any polynomials can be tested, independent of
	   degree and skewness. The score assigned is
	   directly comparable to that of any other polynomials 
	   given to this routine. */

	double size_score;
	double root_score;
	double combined_score;

	if (analyze_poly_size(config, rpoly, apoly, &size_score))
		return;

	root_score = analyze_poly_roots(config, rpoly) +
	             analyze_poly_roots(config, apoly);

	/* Another interesting problem: you have a root score and
	   a size score for each polynomial; how do you combine
	   them to get an overall rating? The value of root_score
	   implies that values of A(x) effectively have size
	   A(x) * exp(root_score). The second term is constant, so
	   pulling this expression out of the integrand used to 
	   calculate the size above gives: */

	combined_score = size_score * pow(exp(root_score), -2.0 /
					(rpoly->degree + apoly->degree));

	/* save the polynomial if its combined score is the 
	   best seen so far. A really thorough implementation
	   would do this in two stages: the best X polynomials
	   would all be kept in stage 1, and in stage 2 the 
	   complete list (possibly combined from the output of
	   different machines) would get a more accurate 
	   root_score computed for each polynomial. The top
	   few such polynomials would then be subjected to 
	   sieving experiments */

	if (combined_score > config->heap[0]->combined_score) {
		poly_select_t *entry = config->heap[0];

		entry->size_score = size_score;
		entry->root_score = root_score;
		entry->combined_score = combined_score;
		memcpy(&entry->rpoly, rpoly, sizeof(mp_poly_t));
		memcpy(&entry->apoly, apoly, sizeof(mp_poly_t));
	}
}

/*------------------------------------------------------------------*/
void analyze_one_poly(msieve_obj *obj,
		      mp_poly_t *rpoly, mp_poly_t *apoly) {

	uint32 i;
	poly_config_t config;

	poly_config_init(obj, &config);
	analyze_poly(&config, rpoly, apoly);

	for (i = 0; i <= rpoly->degree; i++) {
		logprintf(obj, "R%u: %c%s\n", i,
			(rpoly->coeff[i].sign == POSITIVE? ' ' : '-'),
			mp_sprintf(&rpoly->coeff[i].num, 10,
						obj->mp_sprintf_buf));
	}
	for (i = 0; i <= apoly->degree; i++) {
		logprintf(obj, "A%u: %c%s\n", i,
			(apoly->coeff[i].sign == POSITIVE? ' ' : '-'),
			mp_sprintf(&apoly->coeff[i].num, 10,
						obj->mp_sprintf_buf));
	}
	logprintf(obj, "size score = %e, Murphy alpha = %f, combined = %e\n",
			config.heap[0]->size_score,
			config.heap[0]->root_score,
			config.heap[0]->combined_score);
	poly_config_free(&config);
}

/*------------------------------------------------------------------*/
void poly_config_init(msieve_obj *obj, poly_config_t *config) {

	/* one-time initialization for polynomial search */

	uint32 i, num_primes;

	gsl_set_error_handler_off();

	fill_prime_list(&config->prime_list, 0xffff, PRIME_BOUND);
	num_primes = config->prime_list.num_primes;

	config->heap_num_filled = 0;
	for (i = 0; i < POLY_HEAP_SIZE; i++) {
		config->heap[i] = (poly_select_t *)calloc((size_t)1, 
					sizeof(poly_select_t));
	}

	config->integrate_aux = gsl_integration_workspace_alloc((size_t)1000);

	config->samples = (abpair_t *)calloc((size_t)NUM_POLY_SAMPLES,
						sizeof(abpair_t));

	/* when analyzing root properties, all polynomials
	   are sampled at the same set of points */

	i = 0;
	while (i < NUM_POLY_SAMPLES) {
		int32 a = -10000000 +
			(int32)get_rand(&obj->seed1, &obj->seed2) % 20000000;
		uint32 b = get_rand(&obj->seed1, &obj->seed2) % 500000;

		if (b > 0 && mp_gcd_1((uint32)abs(a), b) == 1) {
			config->samples[i].a = a;
			config->samples[i].b = b;
			i++;
		}
	}
}

/*------------------------------------------------------------------*/
void poly_config_free(poly_config_t *config) {

	uint32 i;

	free(config->prime_list.list);
	gsl_integration_workspace_free(config->integrate_aux);
	for (i = 0; i < POLY_HEAP_SIZE; i++)
		free(config->heap[i]);
	free(config->samples);
}
